@@ -1,10 +1,9 @@
-// $Id: double.cpp,v 1.22 2001/09/21 05:06:07 ericb Exp $
+// $Id: double.cpp,v 1.32 2004/02/17 12:58:44 ericb Exp $
 //
 // This software is subject to the terms of the IBM Jikes Compiler
 // License Agreement available at the following URL:
 // http://ibm.com/developerworks/opensource/jikes.
-// Copyright (C) 1996, 1998, 1999, 2000, 2001 International Business
-// Machines Corporation and others.  All Rights Reserved.
+// Copyright (C) 1996, 2004 IBM Corporation and others.  All Rights Reserved.
 // You must accept the terms of that agreement to use this software.
 //
 //
@@ -25,7 +24,7 @@
 //
 // Developed at SunSoft, a Sun Microsystems, Inc. business.
 // Permission to use, copy, modify, and distribute this
-// software is freely granted, provided that this notice 
+// software is freely granted, provided that this notice
 // is preserved.
 // ====================================================
 //
@@ -60,6 +59,7 @@
 
 #include "double.h"
 #include "long.h"
+#include "code.h"
 
 #ifdef HAVE_JIKES_NAMESPACE
 namespace Jikes { // Open namespace Jikes block
@@ -115,10 +115,8 @@ IEEEfloat::IEEEfloat(i4 a)
         a  = -a; // even works for MIN_INT!
         sign = 1;
     }
-    if (a == 0)
-        *this = POSITIVE_ZERO();
-    else
-        *this = Normalize(sign, FRACT_SIZE, (u4) a);
+    *this = (a == 0) ? POSITIVE_ZERO()
+        : Normalize(sign, FRACT_SIZE, (u4) a);
 #endif // HAVE_IEEE754
 }
 
@@ -155,14 +153,21 @@ IEEEfloat::IEEEfloat(const LongInt &a)
         *this = Normalize(sign, FRACT_SIZE, l.LowWord());
     else
     {
-        int exponent = FRACT_SIZE, round = 0;
+        int exponent = FRACT_SIZE - 1, sticky = 0;
         while (l.HighWord())
         {
-            round |= (l.LowWord() & BYTE_MASK) ? 1 : 0;
+            sticky |= (l.LowWord() & BYTE_MASK) ? 1 : 0;
             l >>= 8;
             exponent += 8;
         }
-        *this = Normalize(sign, exponent, l.LowWord() | round);
+        u4 low = l.LowWord();
+        if ((i4) low < 0)
+        {
+            sticky |= (low & BYTE_MASK) ? 1 : 0;
+            low >>= 8;
+            exponent += 8;
+        }
+        *this = Normalize(sign, exponent, low + low + sticky);
     }
 #endif // HAVE_IEEE754
 }
@@ -183,7 +188,7 @@ IEEEfloat::IEEEfloat(const IEEEdouble &d)
             *this = NEGATIVE_INFINITY();
         else if (d.IsNaN())
             *this = NaN();
-        else 
+        else
         {
             //
             // A regular, normalized number - do work on the parts
@@ -319,6 +324,7 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
     // consume whitespace
     //
     for (s = str; ; s++)
+    {
         switch (*s)
         {
         case U_MINUS:
@@ -340,9 +346,10 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
         default:
             goto break2;
         }
+    }
  break2:
     //
-    // consume leading 0's
+    // Consume leading zeroes.
     //
     if (*s == U_0)
     {
@@ -354,6 +361,109 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
             *this = sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
             return;
         }
+        //
+        // Parse hexadecimal floating point.
+        //
+        else if (*s == U_x || *s == U_X)
+        {
+            i4 fraction = 0;
+            // Exponent adjustment, based on '.' location.
+            int shift = FRACT_SIZE;
+            bool seen_dot = false;
+            while (*++s && *s == U_0); // Consume leading zeroes.
+            c = *s;
+            if (c == U_DOT)
+            {
+                seen_dot = true;
+                while (*++s && *s == U_0)
+                    shift -= 4;
+                c = *s;
+            }
+            if (! c || c == U_p || c == U_P)
+            {
+                *this = sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
+                return;
+            }
+            // To avoid overflow, stop after enough bits have been read.
+            for (i = 0; i < (FRACT_SIZE >> 2) + 2 && c; i++, c = *++s)
+            {
+                if (c == U_DOT)
+                {
+                    if (seen_dot)
+                        break;
+                    c = *++s;
+                    seen_dot = true;
+                }
+                int value;
+                if (Code::IsHexDigit(c))
+                    value = Code::Value(c);
+                else break;
+                if (seen_dot)
+                    shift -= 4;
+                fraction = (fraction << 4) + value;
+            }
+            // Round any remaining bits.
+            bool sticky = false;
+            while (c == U_DOT || Code::IsHexDigit(c))
+            {
+                if (c == U_DOT)
+                {
+                    if (seen_dot)
+                        break;
+                    seen_dot = true;
+                }
+                else
+                {
+                    if (! seen_dot)
+                        shift += 4;
+                    if (c != U_0)
+                        sticky = true;
+                }
+                c = *++s;
+            }
+            assert(fraction != 0);
+            if (sticky)
+                fraction++;
+            // On to the expononet.
+            int exponent = 0;
+            esign = false;
+            if (c == U_p || c == U_P)
+            {
+                if (*++s == U_MINUS)
+                {
+                    esign = true;
+                    s++;
+                }
+                else if (*s == U_PLUS)
+                    s++;
+                while ((c = *s++))
+                {
+                    if (! Code::IsDecimalDigit(c))
+                        break;
+                    exponent = exponent * 10 + c - U_0;
+                    // Check for exponent overflow
+                    if (exponent + shift > 19999)
+                    {
+                        if (check_invalid)
+                            *this = NaN();
+                        else
+                        {
+                            *this = esign ? POSITIVE_ZERO()
+                                : POSITIVE_INFINITY();
+                            if (sign)
+                                *this = - *this;
+                        }
+                        return;
+                    }
+                }
+            }
+            if (esign)
+                exponent = - exponent;
+            *this = Normalize(sign, exponent + shift, fraction);
+            if (check_invalid && (IsZero() || IsInfinite()))
+                *this = NaN();
+            return;
+        }
     }
 
     //
@@ -361,7 +471,7 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
     //
     s0 = s;
     y = z = 0;
-    for (nd = nf = 0; (c = *s) >= U_0 && c <= U_9; nd++, s++)
+    for (nd = nf = 0; Code::IsDecimalDigit(c = *s); nd++, s++)
         if (nd < 8)
             y = 10 * y + c - U_0;
     nd0 = nd;
@@ -382,7 +492,7 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
                 nz = 0;
             }
         }
-        for ( ; c >= U_0 && c <= U_9; c = *++s)
+        for ( ; Code::IsDecimalDigit(c); c = *++s)
         {
             nz++;
             if (c -= U_0)
@@ -412,7 +522,7 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
         case U_PLUS:
             c = *++s;
         }
-        if (c >= U_0 && c <= U_9)
+        if (Code::IsDecimalDigit(c))
         {
             if (!nd && !nz && !nz0)
             {
@@ -425,13 +535,13 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
             {
                 L = c - U_0;
                 s1 = s;
-                while ((c = *++s) >= U_0 && c <= U_9)
+                while (Code::IsDecimalDigit(c = *++s))
                     L = 10 * L + c - U_0;
                 //
                 // Avoid confusion from exponents so large that e might
                 // overflow
                 if (s - s1 > 8 || L > 19999)
-                    e = 19999;  
+                    e = 19999;
                 else
                     e = L;
                 if (esign)
@@ -482,7 +592,7 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
     }
     e1 = e -= nf;
 
-    // 
+    //
     // Now we have nd0 digits, starting at s0, followed by a
     // decimal point, followed by nd-nd0 digits.  The number we're
     // after is the integer represented by those digits times 10**e
@@ -529,7 +639,7 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
     }
     e1 += nd - k;
 
-    // 
+    //
     // Get starting approximation: *this * 10**e1
     //
     if (e1 > 0)
@@ -610,7 +720,7 @@ IEEEfloat::IEEEfloat(const char *str, bool check_invalid)
     //
     // Now the hard part -- adjusting *this to the correct value.
     // Put digits into bd: true value = bd * 10^e
-    // 
+    //
     BigInt bd0(s0, nd0, nd, y, 8);
     if (roundup)
         ++bd0;
@@ -710,12 +820,12 @@ i4 IEEEfloat::IntValue() const
 {
     if (IsNaN())
         return 0;
-            
+
     int sign = Sign(),
         exponent = Exponent();
 
-    if (IsInfinite())
-        return sign ? MIN_INT : MAX_INT;
+    if (exponent > 30)
+        return sign ? Int::MIN_INT() : Int::MAX_INT();
 
     // This covers true zero and denorms.
     if (exponent < 0)
@@ -735,11 +845,11 @@ LongInt IEEEfloat::LongValue() const
 {
     if (IsNaN())
         return LongInt(0);
-            
+
     int sign = Sign(),
         exponent = Exponent();
 
-    if (IsInfinite())
+    if (exponent > 62)
         return sign ? LongInt::MIN_LONG() : LongInt::MAX_LONG();
 
     // This covers true zero and denorms.
@@ -760,7 +870,7 @@ IEEEfloat IEEEfloat::Normalize(int sign, int exponent, u4 fraction)
 {
     bool round = false, sticky = false;
 
-    assert(fraction);
+    assert(fraction != 0);
 
     //
     // Normalize right. Shift until value < MAX_FRACT.
@@ -804,10 +914,12 @@ IEEEfloat IEEEfloat::Normalize(int sign, int exponent, u4 fraction)
         return sign ? NEGATIVE_INFINITY() : POSITIVE_INFINITY();
 
     //
-    // Check and respond to underflow 
+    // Check and respond to underflow
     //
     if (exponent <= -BIAS)
     {
+        if (exponent < -BIAS - FRACT_SIZE)
+            return sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
         while (exponent <= -BIAS)
         {
             sticky |= round;
@@ -815,12 +927,9 @@ IEEEfloat IEEEfloat::Normalize(int sign, int exponent, u4 fraction)
             fraction >>= 1;
             exponent++;
         }
-
         if (round && (sticky || (fraction & 1)))
             fraction++;
-
         exponent = -BIAS;
-
         if (fraction == 0)
             return sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
     }
@@ -855,7 +964,7 @@ IEEEfloat IEEEfloat::Ulp() const
     i4 L;
     IEEEfloat f;
     f.value.float_value = value.float_value;
-    
+
     L = (i4) f.ExpBits() - FRACT_SIZE * MIN_FRACT;
     if (L > 0)
         f.value.iword = L;
@@ -887,9 +996,9 @@ bool IEEEfloat::operator== (const IEEEfloat op) const
 #ifdef HAVE_IEEE754
     return value.float_value == op.value.float_value;
 #else
-    return IsNaN() || op.IsNaN() ? false
-                                 : IsZero() && op.IsZero() ? true
-                                                           : value.word == op.value.word;
+    return (IsNaN() || op.IsNaN() ? false
+            : IsZero() && op.IsZero() ? true
+            : value.word == op.value.word);
 #endif
 }
 
@@ -899,7 +1008,7 @@ bool IEEEfloat::operator!= (const IEEEfloat op) const
 #ifdef HAVE_IEEE754
     return value.float_value != op.value.float_value;
 #else
-    return !(*this == op);
+    return ! (*this == op);
 #endif
 }
 
@@ -911,12 +1020,15 @@ bool IEEEfloat::operator< (const IEEEfloat op) const
 #else
     if (IsNaN() || op.IsNaN())
         return false; // NaNs are unordered
-    if (IsZero())
-        return op.IsZero() ? false : op.IsPositive();
-    if (op.IsZero())
-        return IsNegative();
-    // Exploit fact that remaining IEEE floating point sort as signed ints
-    return value.iword < op.value.iword;
+    if (IsZero() && op.IsZero())
+        return false;
+    // Exploit fact that all other IEEE floating point numbers sort like
+    // ints after worrying about sign.
+    if (IsNegative())
+        return op.IsPositive() ||
+            (value.word & ABS_BITS) > (op.value.word & ABS_BITS);
+    return op.IsPositive() &&
+        (value.word & ABS_BITS) < (op.value.word & ABS_BITS);
 #endif
 }
 
@@ -938,12 +1050,15 @@ bool IEEEfloat::operator> (const IEEEfloat op) const
 #else
     if (IsNaN() || op.IsNaN())
         return false; // NaNs are unordered.
-    if (IsZero())
-        return op.IsZero() ? false : op.IsNegative();
-    if (op.IsZero())
-        return IsPositive();
-    // Exploit fact that remaining IEEE floating point sort as signed ints
-    return value.iword > op.value.iword;
+    if (IsZero() && op.IsZero())
+        return false;
+    // Exploit fact that all other IEEE floating point numbers sort like
+    // ints after worrying about sign.
+    if (IsPositive())
+        return op.IsNegative() ||
+            (value.word & ABS_BITS) > (op.value.word & ABS_BITS);
+    return op.IsNegative() &&
+        (value.word & ABS_BITS) < (op.value.word & ABS_BITS);
 #endif
 }
 
@@ -994,7 +1109,7 @@ IEEEfloat IEEEfloat::operator+ (const IEEEfloat op) const
 
     expx = SplitInto((u4 &) x);
     expy = op.SplitInto((u4 &) y);
-    
+
     signx = Sign();
     signy = op.Sign();
 
@@ -1008,7 +1123,7 @@ IEEEfloat IEEEfloat::operator+ (const IEEEfloat op) const
     // Denormalize the fractions, so that the exponents are
     // the same and then set the exponent for the result.
     // Leave enough space for overflow and INT_MIN avoidance!
-    // 
+    //
     if (signx)
         x = -x;
     if (signy)
@@ -1118,7 +1233,7 @@ IEEEfloat IEEEfloat::operator* (const IEEEfloat op) const
     b = a & 0xfffff;
     a >>= 20;
     x = a.LowWord() | ((b > 0) ? 1 : 0);
-    
+
     return Normalize(sign, exponent - 3, x);
 #endif // HAVE_IEEE754
 }
@@ -1142,8 +1257,8 @@ IEEEfloat IEEEfloat::operator/ (const IEEEfloat op) const
     //
 
     if (IsInfinite())
-        return op.IsInfinite() ? NaN()
-                               : sign ? NEGATIVE_INFINITY() : POSITIVE_INFINITY();
+        return (op.IsInfinite() ? NaN()
+                : sign ? NEGATIVE_INFINITY() : POSITIVE_INFINITY());
 
     if (op.IsInfinite())
         return sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
@@ -1200,7 +1315,8 @@ IEEEfloat IEEEfloat::operator% (const IEEEfloat op) const
 {
 #ifdef HAVE_IEEE754
     return IEEEfloat((op.IsZero() ? NaN().value.float_value
-                                  : (float) fmod((double) value.float_value, (double) op.value.float_value)));
+                      : (float) fmod((double) value.float_value,
+                                     (double) op.value.float_value)));
 #else // HAVE_IEEE754
     if (IsNaN() || op.IsNaN())
         return NaN(); // arithmetic on NaNs not allowed
@@ -1301,10 +1417,8 @@ IEEEdouble::IEEEdouble(i4 a)
         a  = -a; // even works for MIN_INT!
         sign = 1;
     }
-    if (a == 0)
-        *this = POSITIVE_ZERO();
-    else
-        *this = Normalize(sign, FRACT_SIZE, ULongInt((u4) a));
+    *this = (a == 0) ? POSITIVE_ZERO()
+        : Normalize(sign, FRACT_SIZE, ULongInt((u4) a));
 #endif // HAVE_IEEE754
 }
 
@@ -1314,8 +1428,8 @@ IEEEdouble::IEEEdouble(const LongInt &a)
 # ifdef HAVE_64BIT_TYPES
     value.double_value = (double)(i8) a.Words();
 # else
-    value.double_value = ((double)(i4) a.HighWord() * (double) 0x40000000 * 4.0) +
-        (double) a.LowWord();
+    value.double_value = ((double)(i4) a.HighWord() * (double) 0x40000000 *
+                          4.0) + (double) a.LowWord();
 # endif // HAVE_64BIT_TYPES
 #else
     int sign = 0;
@@ -1326,10 +1440,8 @@ IEEEdouble::IEEEdouble(const LongInt &a)
         l  = -a; // even works for MIN_LONG!
         sign = 1;
     }
-    if (l == 0)
-        *this = POSITIVE_ZERO();
-    else
-        *this = Normalize(sign, FRACT_SIZE, ULongInt(l));
+    *this = (l == 0) ? POSITIVE_ZERO()
+        : Normalize(sign, FRACT_SIZE, ULongInt(l));
 #endif // HAVE_IEEE754
 }
 
@@ -1359,6 +1471,7 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
     // consume whitespace
     //
     for (s = str; ; s++)
+    {
         switch (*s)
         {
         case U_MINUS:
@@ -1380,9 +1493,10 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
         default:
             goto break2;
         }
+    }
  break2:
     //
-    // consume leading 0's
+    // Consume leading zeroes.
     //
     if (*s == U_0)
     {
@@ -1394,6 +1508,109 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
             *this = sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
             return;
         }
+        //
+        // Parse hexadecimal floating point.
+        //
+        else if (*s == U_x || *s == U_X)
+        {
+            LongInt fraction = 0;
+            // Exponent adjustment, based on '.' location.
+            int shift = FRACT_SIZE;
+            bool seen_dot = false;
+            while (*++s && *s == U_0); // Consume leading zeroes.
+            c = *s;
+            if (c == U_DOT)
+            {
+                seen_dot = true;
+                while (*++s && *s == U_0)
+                    shift -= 4;
+                c = *s;
+            }
+            if (! c || c == U_p || c == U_P)
+            {
+                *this = sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
+                return;
+            }
+            // To avoid overflow, stop after enough bits have been read.
+            for (i = 0; i < (FRACT_SIZE >> 2) + 2 && c; i++, c = *++s)
+            {
+                if (c == U_DOT)
+                {
+                    if (seen_dot)
+                        break;
+                    c = *++s;
+                    seen_dot = true;
+                }
+                int value;
+                if (Code::IsHexDigit(c))
+                    value = Code::Value(c);
+                else break;
+                if (seen_dot)
+                    shift -= 4;
+                fraction = (fraction << 4) + value;
+            }
+            // Round any remaining bits.
+            bool sticky = false;
+            while (c == U_DOT || Code::IsHexDigit(c))
+            {
+                if (c == U_DOT)
+                {
+                    if (seen_dot)
+                        break;
+                    seen_dot = true;
+                }
+                else
+                {
+                    if (! seen_dot)
+                        shift += 4;
+                    if (c != U_0)
+                        sticky = true;
+                }
+                c = *++s;
+            }
+            assert(fraction != 0);
+            if (sticky)
+                fraction++;
+            // On to the expononet.
+            int exponent = 0;
+            esign = false;
+            if (c == U_p || c == U_P)
+            {
+                if (*++s == U_MINUS)
+                {
+                    esign = true;
+                    s++;
+                }
+                else if (*s == U_PLUS)
+                    s++;
+                while ((c = *s++))
+                {
+                    if (! Code::IsDecimalDigit(c))
+                        break;
+                    exponent = exponent * 10 + c - U_0;
+                    // Check for exponent overflow
+                    if (exponent + shift > 19999)
+                    {
+                        if (check_invalid)
+                            *this = NaN();
+                        else
+                        {
+                            *this = esign ? POSITIVE_ZERO()
+                                : POSITIVE_INFINITY();
+                            if (sign)
+                                *this = - *this;
+                        }
+                        return;
+                    }
+                }
+            }
+            if (esign)
+                exponent = - exponent;
+            *this = Normalize(sign, exponent + shift, fraction);
+            if (check_invalid && (IsZero() || IsInfinite()))
+                *this = NaN();
+            return;
+        }
     }
 
     //
@@ -1401,7 +1618,7 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
     //
     s0 = s;
     y = z = 0;
-    for (nd = nf = 0; (c = *s) >= U_0 && c <= U_9; nd++, s++)
+    for (nd = nf = 0; Code::IsDecimalDigit(c = *s); nd++, s++)
         if (nd < 9)
             y = 10 * y + c - U_0;
         else if (nd < 16)
@@ -1424,7 +1641,7 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
                 nz = 0;
             }
         }
-        for ( ; c >= U_0 && c <= U_9; c = *++s)
+        for ( ; Code::IsDecimalDigit(c); c = *++s)
         {
             nz++;
             if (c -= U_0)
@@ -1458,7 +1675,7 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
         case U_PLUS:
             c = *++s;
         }
-        if (c >= U_0 && c <= U_9)
+        if (Code::IsDecimalDigit(c))
         {
             if (!nd && !nz && !nz0)
             {
@@ -1471,13 +1688,13 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
             {
                 L = c - U_0;
                 s1 = s;
-                while ((c = *++s) >= U_0 && c <= U_9)
+                while (Code::IsDecimalDigit(c = *++s))
                     L = 10 * L + c - U_0;
                 //
                 // Avoid confusion from exponents so large that e might
                 // overflow
                 if (s - s1 > 8 || L > 19999)
-                    e = 19999;  
+                    e = 19999;
                 else
                     e = L;
                 if (esign)
@@ -1528,7 +1745,7 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
     }
     e1 = e -= nf;
 
-    // 
+    //
     // Now we have nd0 digits, starting at s0, followed by a
     // decimal point, followed by nd-nd0 digits.  The number we're
     // after is the integer represented by those digits times 10**e
@@ -1577,7 +1794,7 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
     }
     e1 += nd - k;
 
-    // 
+    //
     // Get starting approximation: *this * 10**e1
     //
     if (e1 > 0)
@@ -1658,7 +1875,7 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
     //
     // Now the hard part -- adjusting *this to the correct value.
     // Put digits into bd: true value = bd * 10^e
-    // 
+    //
     BigInt bd0(s0, nd0, nd, y, 9);
     if (roundup)
         ++bd0;
@@ -1793,7 +2010,8 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
             *this += aadj1 * Ulp();
             if (Exponent() >= BIAS - FRACT_SIZE)
             {
-                if (tmp.HighWord() == POS_INF_HI - 1 && tmp.LowWord() == ZERO_LO - 1)
+                if (tmp.HighWord() == POS_INF_HI - 1 &&
+                    tmp.LowWord() == ZERO_LO - 1)
                 {
                     // overflow
                     *this = POSITIVE_INFINITY();
@@ -1846,20 +2064,20 @@ IEEEdouble::IEEEdouble(const char *str, bool check_invalid)
 i4 IEEEdouble::IntValue() const
 {
     if (IsNaN())
-        return 0;                                                             
-            
+        return 0;
+
 #ifdef HAVE_IEEE754
-    if (value.double_value < (double)(i4) MIN_INT)
-        return MIN_INT;
-    else if (value.double_value > (double) MAX_INT)
-        return MAX_INT;
+    if (value.double_value < (double)(i4) Int::MIN_INT())
+        return Int::MIN_INT();
+    else if (value.double_value > (double) Int::MAX_INT())
+        return Int::MAX_INT();
     return (i4) value.double_value;
 #else
     int sign = Sign(),
         exponent = Exponent();
 
-    if (IsInfinite() || exponent > 30)
-        return sign ? MIN_INT : MAX_INT;
+    if (exponent > 30)
+        return sign ? Int::MIN_INT() : Int::MAX_INT();
 
     // This includes true zero and denorms.
     if (exponent < 0)
@@ -1868,18 +2086,18 @@ i4 IEEEdouble::IntValue() const
     i4 result = (i4) (Fraction() >> (FRACT_SIZE - exponent)).LowWord();
 
     return sign ? -result : result;
-#endif // HAVE_IEEE754
+#endif // ! HAVE_IEEE754
 }
 
 LongInt IEEEdouble::LongValue() const
 {
     if (IsNaN())
         return LongInt(0);
-            
+
     int sign = Sign(),
         exponent = Exponent();
 
-    if (IsInfinite() || exponent > 62)
+    if (exponent > 62)
         return sign ? LongInt::MIN_LONG() : LongInt::MAX_LONG();
 
     // This covers true zero and denorms.
@@ -1914,7 +2132,9 @@ IEEEdouble IEEEdouble::Normalize(int sign, int exponent, ULongInt fraction)
             fraction >>= 1;
             exponent++;
         }
-        if (round && (sticky || (fraction.LowWord() & 1)) && exponent > -(int) BIAS)
+        if (round && (sticky || (fraction.LowWord() & 1)) &&
+            exponent > -(int) BIAS)
+        {
             //
             // Capture any overflow caused by rounding. No other checks are
             // required because if overflow occurred, the the low order bit
@@ -1925,6 +2145,7 @@ IEEEdouble IEEEdouble::Normalize(int sign, int exponent, ULongInt fraction)
                 fraction >>= 1;
                 exponent++;
             }
+        }
     }
 
     //
@@ -1940,27 +2161,26 @@ IEEEdouble IEEEdouble::Normalize(int sign, int exponent, ULongInt fraction)
     //
     // Check and respond to overflow
     //
-    if (exponent > (int) BIAS)
+    if (exponent > BIAS)
         return sign ? NEGATIVE_INFINITY() : POSITIVE_INFINITY();
 
     //
-    // Check and respond to underflow 
+    // Check and respond to underflow
     //
-    if (exponent <= -(int) BIAS)
+    if (exponent <= -BIAS)
     {
-        while (exponent <= -(int) BIAS)
+        if (exponent < -BIAS - FRACT_SIZE)
+            return sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
+        while (exponent <= -BIAS)
         {
             sticky |= round;
             round = (fraction.LowWord() & 1) != 0;
             fraction >>= 1;
             exponent++;
         }
-
         if (round && (sticky || (fraction.LowWord() & 1)))
             fraction++;
-
-        exponent = -(int) BIAS;
-
+        exponent = -BIAS;
         if (fraction == 0)
             return sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
     }
@@ -1993,7 +2213,7 @@ IEEEdouble IEEEdouble::Ulp() const
     i4 L;
     IEEEdouble d;
     d.value.double_value = value.double_value;
-    
+
     L = (i4) d.ExpBits() - FRACT_SIZE * MIN_FRACT;
     if (L > 0)
         d.setHighAndLowWords((u4) L, 0);
@@ -2005,7 +2225,7 @@ IEEEdouble IEEEdouble::Ulp() const
         else
         {
             L -= FRACT_SIZE_HI;
-            d.setHighAndLowWords(0, L >= 31 ? 1 : 1 << 31 - L);
+            d.setHighAndLowWords(0, L >= 31 ? 1 : 1 << (31 - L));
         }
     }
     return d;
@@ -2032,9 +2252,9 @@ bool IEEEdouble::operator== (const IEEEdouble op) const
     // TODO: Microsoft VC++ botches this, mixing 12.0 and NaN
     return value.double_value == op.value.double_value;
 #else
-    return IsNaN() || op.IsNaN() ? false
-                                 : IsZero() && op.IsZero() ? true
-                                                           : (BaseLong) *this == (BaseLong) op;
+    return (IsNaN() || op.IsNaN() ? false
+            : IsZero() && op.IsZero() ? true
+            : (BaseLong) *this == (BaseLong) op);
 #endif
 }
 
@@ -2082,10 +2302,10 @@ IEEEdouble IEEEdouble::operator+ (const IEEEdouble op) const
     //
     LongInt x, y, round = 0;
     int expx, expy, signx, signy;
-    
+
     expx = SplitInto(x);
     expy = op.SplitInto(y);
-    
+
     signx = Sign();
     signy = op.Sign();
 
@@ -2099,7 +2319,7 @@ IEEEdouble IEEEdouble::operator+ (const IEEEdouble op) const
     // Denormalize the fractions, so that the exponents are
     // the same and then set the exponent for the result.
     // Leave enough space for overflow and LONG_MIN avoidance!
-    // 
+    //
     if (signx)
         x = -x;
     if (signy)
@@ -2226,10 +2446,10 @@ IEEEdouble IEEEdouble::operator/ (const IEEEdouble op) const
 {
 #ifdef HAVE_IEEE754
     return op.IsZero() ? ((IsNaN() || IsZero()) ? NaN()
-                               : (IsPositive() ^ op.IsPositive())
-                                       ? NEGATIVE_INFINITY()
-                                       : POSITIVE_INFINITY())
-                       : IEEEdouble(value.double_value / op.value.double_value);
+                          : ((IsPositive() ^ op.IsPositive())
+                             ? NEGATIVE_INFINITY()
+                             : POSITIVE_INFINITY()))
+        : IEEEdouble(value.double_value / op.value.double_value);
 #else // HAVE_IEEE754
     if (IsNaN() || op.IsNaN())
         return NaN(); // arithmetic on NaNs not allowed
@@ -2241,8 +2461,8 @@ IEEEdouble IEEEdouble::operator/ (const IEEEdouble op) const
     //
 
     if (IsInfinite())
-        return op.IsInfinite() ? NaN()
-                               : sign ? NEGATIVE_INFINITY() : POSITIVE_INFINITY();
+        return (op.IsInfinite() ? NaN()
+                : sign ? NEGATIVE_INFINITY() : POSITIVE_INFINITY());
 
     if (op.IsInfinite())
         return sign ? NEGATIVE_ZERO() : POSITIVE_ZERO();
@@ -2299,7 +2519,7 @@ IEEEdouble IEEEdouble::operator% (const IEEEdouble op) const
 {
 #ifdef HAVE_IEEE754
     return IEEEdouble((op.IsZero() ? NaN().value.double_value
-                                   : fmod(value.double_value, op.value.double_value)));
+                       : fmod(value.double_value, op.value.double_value)));
 #else // HAVE_IEEE754
     if (IsNaN() || op.IsNaN())
         return NaN(); // arithmetic on NaNs not allowed
@@ -2358,13 +2578,16 @@ bool IEEEdouble::operator< (const IEEEdouble op) const
 #else
     if (IsNaN() || op.IsNaN())
         return false; // NaNs are unordered
-    if (IsZero())
-        return op.IsZero() ? false : op.IsPositive();
-    if (op.IsZero())
-        return IsNegative();
-    // Exploit fact that remaining IEEE floating point sort as signed ints
-    u4 a = HighWord(), b = op.HighWord();
-    return (a < b) || ((a == b) && LowWord() < op.LowWord());
+    if (IsZero() && op.IsZero())
+        return false;
+    // Exploit fact that all other IEEE floating point numbers sort like
+    // ints after worrying about sign.
+    u4 a = HighWord() & ABS_BITS, b = op.HighWord() & ABS_BITS;
+    if (IsNegative())
+        return op.IsPositive() ||
+            (a > b || (a == b && LowWord() > op.LowWord()));
+    return op.IsPositive() &&
+        (a < b || (a == b && LowWord() < op.LowWord()));
 #endif
 }
 
@@ -2386,13 +2609,16 @@ bool IEEEdouble::operator> (const IEEEdouble op) const
 #else
     if (IsNaN() || op.IsNaN())
         return false; // NaNs are unordered.
-    if (IsZero())
-        return op.IsZero() ? false : op.IsNegative();
-    if (op.IsZero())
-        return IsPositive();
-    // Exploit fact that remaining IEEE floating point sort as signed ints
-    u4 a = HighWord(), b = op.HighWord();
-    return (a > b) || ((a == b) && LowWord() > op.LowWord());
+    if (IsZero() && op.IsZero())
+        return false;
+    // Exploit fact that all other IEEE floating point numbers sort like
+    // ints after worrying about sign.
+    u4 a = HighWord() & ABS_BITS, b = op.HighWord() & ABS_BITS;
+    if (IsPositive())
+        return op.IsNegative() ||
+            (a > b || (a == b && LowWord() > op.LowWord()));
+    return op.IsNegative() &&
+        (a < b || (a == b && LowWord() < op.LowWord()));
 #endif
 }
 
@@ -2459,7 +2685,7 @@ BigInt::BigInt(const IEEEdouble &d, int &e, int &bits) : data(NULL)
     {
         if ((k = lo0bits(y)) != 0)
         {
-            data[0] = y | z << 32 - k;
+            data[0] = y | z << (32 - k);
             z >>= k;
         }
         else
@@ -2607,7 +2833,7 @@ BigInt &BigInt::operator +(const unsigned op) const
     ULongInt sum; // sum
     BigInt *result = new BigInt(*this);
     u4 *x = result -> data; // access to data
-    
+
     do
     {
         sum = ULongInt(*x) + carry;
@@ -2686,7 +2912,7 @@ BigInt &BigInt::operator *(unsigned op) const
     ULongInt product; // product
     BigInt *result = new BigInt(*this);
     u4 *x = result -> data; // access to data
-    ULongInt factor = op; // avoid creating object multiple times
+    ULongInt factor = (u4) op; // avoid creating object multiple times
 
     do
     {
@@ -2807,8 +3033,8 @@ BigInt &BigInt::multadd(unsigned m, unsigned a)
     u4 *x = data; // access to data
     u4 carry = a; // carry between words
     ULongInt product; // product
-    ULongInt factor = m; // avoid creating object multiple times
-    
+    ULongInt factor = (u4) m; // avoid creating object multiple times
+
     do
     {
         product = ULongInt(*x) * factor + carry;
@@ -2940,10 +3166,10 @@ IEEEfloat BigInt::FloatValue() const
     y = *--xa;
     k = hi0bits(y);
     if (k < 8)
-        return IEEEfloat(0x3f800000 | y >> 8 - k);
+        return IEEEfloat(0x3f800000 | y >> (8 - k));
     z = xa > data ? *--xa : 0;
     if (k -= 8)
-        return IEEEfloat(0x3f800000 | y << k | z >> 32 - k);
+        return IEEEfloat(0x3f800000 | y << k | z >> (32 - k));
     else
         return IEEEfloat(0x3f800000 | y);
 }
@@ -2957,17 +3183,17 @@ IEEEdouble BigInt::DoubleValue() const
     k = hi0bits(y);
     if (k < 11)
     {
-        hi = 0x3ff00000 | y >> 11 - k;
+        hi = 0x3ff00000 | y >> (11 - k);
         w = xa > data ? *--xa : 0;
-        lo = y << (32- 11) + k | w >> 11 - k;
+        lo = y << (32 - 11 + k) | w >> (11 - k);
         return IEEEdouble(hi, lo);
     }
     z = xa > data ? *--xa : 0;
     if (k -= 11)
     {
-        hi = 0x3ff00000 | y << k | z >> 32 - k;
+        hi = 0x3ff00000 | y << k | z >> (32 - k);
         y = xa > data ? *--xa : 0;
-        lo = z << k | y >> 32 - k;
+        lo = z << k | y >> (32 - k);
     }
     else
     {
