@@ -1,27 +1,20 @@
-// $Id: tab.h,v 1.9 2001/01/05 09:13:21 mdejong Exp $
+// $Id: tab.h,v 1.11 2001/09/14 05:31:34 ericb Exp $ -*- c++ -*-
 //
 // This software is subject to the terms of the IBM Jikes Compiler
 // License Agreement available at the following URL:
-// http://www.ibm.com/research/jikes.
-// Copyright (C) 1996, 1998, International Business Machines Corporation
-// and others.  All Rights Reserved.
+// http://ibm.com/developerworks/opensource/jikes.
+// Copyright (C) 1996, 1998, 1999, 2000, 2001 International Business
+// Machines Corporation and others.  All Rights Reserved.
 // You must accept the terms of that agreement to use this software.
 //
+
 #ifndef tab_INCLUDED
 #define tab_INCLUDED
 
 #include "platform.h"
 
-/*
-//FIXME: include stuff
-#ifdef HAVE_WCHAR_H
-# include <wchar.h>
-#endif
-#include <string.h>
-*/
-
-#ifdef	HAVE_JIKES_NAMESPACE
-namespace Jikes {	// Open namespace Jikes block
+#ifdef HAVE_JIKES_NAMESPACE
+namespace Jikes { // Open namespace Jikes block
 #endif
 
 class Tab
@@ -38,9 +31,9 @@ private:
     static int tab_size;
 };
 
-#ifdef	HAVE_JIKES_NAMESPACE
-}			// Close namespace Jikes block
+#ifdef HAVE_JIKES_NAMESPACE
+} // Close namespace Jikes block
 #endif
 
-#endif
+#endif // tab_INCLUDED
 
